﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace NesExtendPatchTool
{

	#region IpsException

	public class IpsException : Exception
	{
		public IpsException()
		{
		}

		public IpsException(string str)
			: base(str)
		{
		}
	}

	#endregion IpsException

	#region PatchInfo

	/// <summary>
	/// パッチの1単位の情報。
	/// </summary>
	internal class PatchInfo
	{

		/// <summary>
		/// この情報(パッチの1単位)のサイズ。
		/// </summary>
		public int InfoSize { get; set; }

		/// <summary>
		/// パッチ先アドレス。
		/// </summary>
		public int Addr { get; set; }

		/// <summary>
		/// 書き込むデータ。
		/// </summary>
		public byte[] Data { get; set; }

		/// <summary>
		/// 書き込むデータのサイズ。
		/// </summary>
		public int Size
		{
			get
			{
				if (Data != null)
				{
					return Data.Length;
				}
				else
				{
					return 0;
				}
			}
		}

		/// <summary>
		/// パッチ情報をIPSデータの指定アドレスから作成。(読み込む)
		/// </summary>
		/// <param name="patchData">IPSデータ</param>
		/// <param name="srcAddr">アドレス</param>
		public PatchInfo(byte[] ips, int srcAddr)
		{
			int addr = srcAddr;

			// アドレス取得
			this.Addr = ReadInteger3Byte(ips, addr);
			addr += 3;
			// サイズ取得
			int size = ReadInteger2Byte(ips, addr);
			addr += 2;
			// フィル判定
			if (size == 0)
			{
				// フィル時は埋めるサイズを取得
				size = ReadInteger2Byte(ips, addr);
				addr += 2;
				// フィル値を取得
				byte fill = ips[addr];
				addr += 1;
				// データ作成
				this.Data = new byte[size];
				for (int i = 0; i < this.Data.Length; i++)
				{
					this.Data[i] = fill;
				}
				// 単位サイズを設定
				this.InfoSize = addr - srcAddr;
			}
			else
			{
				// 通常時はデータを取得
				this.Data = new byte[size];
				for (int i = 0; i < this.Data.Length; i++)
				{
					this.Data[i] = ips[addr];
					addr++;
				}
				// 単位サイズを設定
				this.InfoSize = addr - srcAddr;
			}
		}

		/// <summary>
		/// データから2バイト読み込み
		/// </summary>
		/// <param name="data">データ</param>
		/// <param name="addr">アドレス</param>
		/// <returns>値</returns>
		private int ReadInteger2Byte(byte[] data, int addr)
		{
			int result = 0;

			result = (data[addr] << 8) + data[addr + 1];

			return result;
		}

		/// <summary>
		/// データから3バイト読み込み
		/// </summary>
		/// <param name="data">データ</param>
		/// <param name="addr">アドレス</param>
		/// <returns>値</returns>
		private int ReadInteger3Byte(byte[] data, int addr)
		{
			int result = 0;

			result = (((data[addr] << 8) + data[addr + 1]) << 8) + data[addr + 2];

			return result;
		}


	}

	#endregion PatchInfo



	/// <summary>
	/// IPSパッチ処理クラス。
	/// ・エラー時は例外発生。
	/// </summary>
	public class Ips
	{
		/// <summary>
		/// パッチ実行。
		/// ・サイズ調整をする
		/// ・実際にパッチデータを書き込む
		/// </summary>
		/// <param name="data">対象データ</param>
		/// <param name="patchData">IPSデータ</param>
		public static void Patch(List<byte> data, byte[] patchData)
		{
			try
			{
				// チェック
				CheckIps(patchData);
				// IPSで書き込む最後のアドレスを取得
				int max = SearchIpsLast(patchData);
				// 対象データが小さい場合、0データを追加する
				if (max > data.Count)
				{
					for (int i = data.Count; i < max; i++)
					{
						data.Add(0x00);
					}
				}
				// パッチ実行
				PatchIps(data, patchData);
			}
			catch (Exception e)
			{
				throw new IpsException(e.Message);
			}
		}

		/// <summary>
		/// 最低限のチェック。
		/// </summary>
		/// <param name="patchData">IPSデータ</param>
		private static void CheckIps(byte[] patchData)
		{
			if (
				(patchData == null) ||
				(patchData.Length < 8) ||
				(patchData[0] != 'P') ||
				(patchData[1] != 'A') ||
				(patchData[2] != 'T') ||
				(patchData[3] != 'C') ||
				(patchData[4] != 'H') ||
				(patchData[patchData.Length - 3] != 'E') ||
				(patchData[patchData.Length - 2] != 'O') ||
				(patchData[patchData.Length - 1] != 'F')
				)
			{
				throw new Exception("Not IPS file.");
			}
		}

		/// <summary>
		/// IPSで書き込む最後のアドレスを取得。
		/// </summary>
		/// <param name="patchData">IPSデータ</param>
		/// <returns>アドレス</returns>
		private static int SearchIpsLast(byte[] patchData)
		{
			// 解析
			int lastAddr = 0;
			for (int i = 5; (i + 5) < patchData.Length; )
			{
				PatchInfo info;
				try
				{
					info = new PatchInfo(patchData, i);
					i += info.InfoSize;
				}
				catch (Exception e)
				{
					throw new IpsException("Reading erro PatchInfo.\n" + e.Message);
				}
				// 最終アドレスチェック
				lastAddr = (info.Addr + info.Size);
			}

			return lastAddr;
		}

		/// <summary>
		/// IPSパッチ。
		/// ・実際に書き換える。
		/// </summary>
		/// <param name="data">対象データ</param>
		/// <param name="patchData">IPSデータ</param>
		private static void PatchIps(List<byte> data, byte[] patchData)
		{
			for (int i = 5; (i + 5) < patchData.Length; )
			{
				PatchInfo info;
				try
				{
					info = new PatchInfo(patchData, i);
					i += info.InfoSize;
				}
				catch (Exception e)
				{
					throw new IpsException("Reading erro PatchInfo.\n" + e.Message);
				}
				// 書き込む
				for (int p = 0; p < info.Size; p++)
				{
					int tAddr = info.Addr + p;
					data[tAddr] = info.Data[p];
				}
			}
		}
	}
}
