﻿using System;

namespace NesExtendPatchTool
{

	/// <summary>
	/// NES Header class.
	/// NESヘッダを扱うクラス。
	///  @ 2010/03/14 Yy
	/// </summary>
	public class NesHeader
	{

		#region enum

		/// <summary>
		/// Flags.
		/// フラグ。
		/// </summary>
		public enum Flags
		{
			MirrorVertical = 1,
			Sram = 2,
			Trainer = 4,
			FourScreen = 8
		}

		/// <summary>
		/// Mirror setting.
		/// ミラー設定。
		/// </summary>
		public enum Mirroring
		{
			Horizontal = 0,
			Vertical = 1
		}

		#endregion enum

		#region Property

		private byte[] mData = null;
		/// <summary>
		/// Loaded data.
		/// 読み込んだデータ。
		/// </summary>
		public byte[] Data
		{
			get
			{
				return mData;
			}
		}

		#region Basic

		/// <summary>
		/// Check data is NES.
		/// NESファイルか判定する。
		/// - Data[0-3]
		/// </summary>
		public bool IsNesHeader
		{
			get
			{
				bool result = false;
				if (Data != null)
				{
					if ((Data[0] == 'N') && (Data[1] == 'E') && (Data[2] == 'S') && (Data[3] == 0x1A))
					{
						result = true;
					}
				}
				return result;
			}
		}

		/// <summary>
		/// PRG-ROM bank count.
		/// PRG-ROMのバンク数（16KB単位）
		/// - Data[4]
		/// </summary>
		public byte PrgRom
		{
			get
			{
				return Data[4];
			}
			set
			{
				Data[4] = value;
			}
		}

		/// <summary>
		/// CHR-ROM bank count.
		/// CHR-ROMのバンク数（8KB単位）
		/// - Data[5]
		/// </summary>
		public byte ChrRom
		{
			get
			{
				return Data[5];
			}
			set
			{
				Data[5] = value;
			}
		}

		/// <summary>
		/// Mapper.
		/// マッパ番号。
		/// - Data[6-7]
		/// </summary>
		public byte Mapper
		{
			get
			{
				int mapperL = (Data[6] & 0xF0) >> 4;
				int mapperH = Data[7] & 0xF0;
				int mapper = mapperH | mapperL;
				return (byte)mapper;
			}
			set
			{
				byte mapperL = (byte)(value & 0x0F);
				byte mapperH = (byte)(value & 0xF0);
				byte flags = (byte)(Data[6] & 0x0F);
				Data[6] = (byte)((mapperL << 4) | flags);
				Data[7] = (byte)mapperH;
			}
		}

		#endregion Basic

		#region Flags

		/// <summary>
		/// Mirror Setting.
		/// ミラー設定。
		/// </summary>
		public Mirroring Mirror
		{
			get
			{
				if ((Data[6] & (byte)Flags.MirrorVertical) == (byte)Flags.MirrorVertical)
				{
					return Mirroring.Horizontal;
				}
				else
				{
					return Mirroring.Vertical;
				}
			}
			set
			{
				Data[6] = (byte)(Data[6] | (byte)Flags.MirrorVertical);
			}
		}

		/// <summary>
		/// Battery backup S-RAM.
		/// バッテリーバックアップ。
		/// </summary>
		public bool IsSRam
		{
			get
			{
				return ((Data[6] & (byte)Flags.Sram) == (byte)Flags.Sram);
			}
			set
			{
				Data[6] = (byte)(Data[6] | (byte)Flags.Sram);
			}
		}

		/// <summary>
		/// Trainer.
		/// トレイナー。
		/// </summary>
		public bool IsTrainer
		{
			get
			{
				return ((Data[6] & (byte)Flags.Trainer) == (byte)Flags.Trainer);
			}
			set
			{
				Data[6] = (byte)(Data[6] | (byte)Flags.Trainer);
			}
		}

		/// <summary>
		/// 4 screen.
		/// 4画面バッファ。
		/// </summary>
		public bool Is4Screen
		{
			get
			{
				return ((Data[6] & (byte)Flags.FourScreen) == (byte)Flags.FourScreen);
			}
			set
			{
				Data[6] = (byte)(Data[6] | (byte)Flags.FourScreen);
			}
		}

		#endregion Flags

		#endregion Property

		/// <summary>
		/// Constructor.
		/// コンストラクタ。
		/// </summary>
		/// <param name="data">NES Data with header</param>
		public NesHeader(byte[] data)
		{
			mData = data;

			if (mData == null)
			{
				throw new Exception("Not NES file.");
			}

			if (!IsNesHeader)
			{
				throw new Exception("Not NES file.");
			}
		}

	}
}
