﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace NesExtendPatchTool
{

	#region enum

	/// <summary>
	/// 処理タイプ。
	/// </summary>
	internal enum ProcessType
	{
		// 何もしない
		None = 0,
		// サイズを確認する
		CheckSize = 1,
		// データを追加する
		Add = 2,
		// ファイルを追加する
		AddFile = 10,
		// IPSパッチを当てる
		Ips = 3,
		// ファイルを指定する
		SetFile = 4,
		// ファイルを書き出す
		WriteFile = 5,
		// ダミーデータを追加する
		AddDummy = 6,
	}

	#endregion enum

	#region NESP ProcessData

	/// <summary>
	/// NESP処理単位。(スクリプトの行イメージ)
	/// ・処理のパラメータをプロパティに読み込む。
	/// </summary>
	internal class NespProcess
	{

		private const string DEFAULT_INITIAL = "@";

		private char[] TRIM_CHAR = { ' ', '"' };
		private char[] SPLIT_CHAR = { ' ', '.', ':' };

		#region Property

		public ProcessType ProcessType { get; set; }

		public string Line { get; set; }
		public string Command { get; set; }
		public int Size { get; set; }
		public byte ByteValue { get; set; }
		public string Filename { get; set; }

		private string mInitial = "";
		/// <summary>
		/// 参照元ファイルの識別記号。(省略可能)
		/// ・A.PRG、B.CHR等のように記述して、複数ファイルを扱うことができる。
		/// ・省略可能で、省略した場合は"@"が使われる。
		/// </summary>
		public string Initial
		{
			get
			{
				string initial = mInitial;
				if (string.IsNullOrEmpty(initial)) initial = DEFAULT_INITIAL;
				return initial;
			}
			set
			{
				mInitial = value;
			}
		}

		/// <summary>
		/// このアドレスはCHRの場合CHR_ADDRを足す必要がある
		/// </summary>
		public int Addr { get; set; }

		#endregion Property

		/// <summary>
		/// 行からNESPインスタンスを作成。
		/// </summary>
		/// <param name="text"></param>
		public NespProcess(string text)
		{
			try
			{
				LoadParameterFromScript(text);
			}
			catch
			{
				throw new NespException(text);
			}
		}

		/// <summary>
		/// 行からパラメータを取得。
		/// </summary>
		/// <param name="text"></param>
		private void LoadParameterFromScript(string text)
		{
			// 空の場合は例外
			if (string.IsNullOrEmpty(text))
			{
				throw new Exception("NESP Script error.");
			}

			this.Line = text.Trim();

			// 大文字変換
			string line = this.Line.ToUpper();
			// スペース分割
			string[] tokenOrg = SplitLine(line, SPLIT_CHAR);

			// 空行
			if ((tokenOrg == null) || (tokenOrg.Length < 1))
			{
				throw new Exception("NESP Script error");
			}

			// イニシャルを取得・トークンを調整
			int tokenStart;
			if (tokenOrg[0].Length == 1)
			{
				// イニシャル有りの場合
				this.Initial = tokenOrg[0].ToUpper();
				tokenStart = 1;
			}
			else
			{
				// デフォルトはイニシャル無し
				this.Initial = "";
				tokenStart = 0;
			}
			int tokenLength = tokenOrg.Length - tokenStart;
			string[] token = new string[tokenLength];
			for (int i = 0; i < tokenLength; i++)
			{
				token[i] = tokenOrg[i + tokenStart];
			}


			// 空行
			if ((token == null) || (token.Length < 1))
			{
				throw new Exception("NESP Script error");
			}

			this.Command = token[0].ToUpper();

			// 判定
			switch (this.Command)
			{

				// Initial非サポートコマンド

				case "IPS":
					{
						ThrowExceptionIfSetInitial(text);
						if (token.Length == 1)
						{
							// IPSファイル名省略の場合
							this.Filename = string.Empty;
							this.ProcessType = ProcessType.Ips;
						}
						if (token.Length == 2)
						{
							if (!string.IsNullOrEmpty(token[1]))
							{
								this.Filename = TrimQuotedString(token[1]);
								this.ProcessType = ProcessType.Ips;
							}
						}
					}
					break;

				case "FILE":
					{
						ThrowExceptionIfSetInitial(text);
						this.Filename = TrimQuotedString(token[1]);
						this.ProcessType = ProcessType.AddFile;
					}
					break;

				case "DUMMY":
					{
						ThrowExceptionIfSetInitial(text);
						if (token.Length == 2)
						{
							this.ByteValue = 0;
							this.Size = Convert.ToInt32(token[1], 16);
						}
						if (token.Length == 3)
						{
							this.ByteValue = Convert.ToByte(token[1], 16);
							this.Size = Convert.ToInt32(token[2], 16);
						}
						this.ProcessType = ProcessType.AddDummy;
					}
					break;

				case "WRITE_FILE":
					{
						ThrowExceptionIfSetInitial(text);
						this.Filename = TrimQuotedString(token[1]);
						this.ProcessType = ProcessType.WriteFile;
					}
					break;

				// Initialサポートコマンド

				case "SET_FILE":
					{
						this.Filename = TrimQuotedString(token[1]);
						this.ProcessType = ProcessType.SetFile;
					}
					break;

				case "CHECK_SIZE":
					{
						this.Size = Convert.ToInt32(token[1], 16);
						this.ProcessType = ProcessType.CheckSize;
					}
					break;

				case "HEAD":
					{
						this.Addr = 0x0000;
						this.Size = 0x0010;	// 16byte
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "PRG":
					{
						this.Size = 0x4000;	// 16KB
						int no = Convert.ToInt32(token[1]);
						this.Addr = this.Size * no;
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "PRG1KB":
					{
						this.Size = 0x0400;	// 1KB
						int no = Convert.ToInt32(token[1]);
						this.Addr = this.Size * no;
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "PRG2KB":
					{
						this.Size = 0x0800;	// 2KB
						int no = Convert.ToInt32(token[1]);
						this.Addr = this.Size * no;
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "PRG4KB":
					{
						this.Size = 0x1000;	// 4KB
						int no = Convert.ToInt32(token[1]);
						this.Addr = this.Size * no;
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "PRG8KB":
					{
						this.Size = 0x2000;	// 8KB
						int no = Convert.ToInt32(token[1]);
						this.Addr = this.Size * no;
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "PRG16KB":
					{
						this.Size = 0x4000;	// 16KB
						int no = Convert.ToInt32(token[1]);
						this.Addr = this.Size * no;
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "CHR":
					{
						this.Size = 0x2000;	// 8KB
						int no = Convert.ToInt32(token[1]);
						this.Addr = this.Size * no;
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "CHR1KB":
					{
						this.Size = 0x0400;	// 1KB
						int no = Convert.ToInt32(token[1]);
						this.Addr = this.Size * no;
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "CHR2KB":
					{
						this.Size = 0x0800;	// 2KB
						int no = Convert.ToInt32(token[1]);
						this.Addr = this.Size * no;
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "CHR4KB":
					{
						this.Size = 0x1000;	// 4KB
						int no = Convert.ToInt32(token[1]);
						this.Addr = this.Size * no;
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "CHR8KB":
					{
						this.Size = 0x2000;	// 8KB
						int no = Convert.ToInt32(token[1]);
						this.Addr = this.Size * no;
						this.ProcessType = ProcessType.Add;
					}
					break;

				case "COPY":
					{
						if (token.Length == 3)
						{
							this.Addr = Convert.ToInt32(token[1], 16);
							this.Size = Convert.ToInt32(token[2], 16);
							this.ProcessType = ProcessType.Add;
						}
					}
					break;

				default:
					{
						ThrowBadCommandException(text);
					}
					break;
			}
		}

		#region private

		/// <summary>
		/// 非対応コマンドが指定された場合の例外を投げる。
		/// </summary>
		/// <param name="text"></param>
		private void ThrowBadCommandException(string text)
		{
			NespException e = new NespException(text);
			e.BadCommand = true;
			throw e;
		}

		/// <summary>
		/// イニシャル非対応コマンドでイニシャルを指定たときの例外を投げる。
		/// </summary>
		/// <param name="text"></param>
		private void ThrowExceptionIfSetInitial(string text)
		{
			if (this.Initial != DEFAULT_INITIAL)
			{
				NespException e = new NespException(text);
				e.BadInitialForCommand = true;
				throw e;
			}
		}

		/// <summary>
		/// テキストを指定したデリミタで分割する。
		/// ・string.Split()のダブルクォート囲み対応版。
		/// </summary>
		/// <param name="line">テキスト</param>
		/// <param name="splitChars">デリミタ</param>
		/// <returns>分割</returns>
		private string[] SplitLine(string line, char[] splitChars)
		{
			List<string> token = new List<string>();
			while (!string.IsNullOrEmpty(line))
			{
				int minIndex = int.MaxValue;
				// ダブルクォートの場合、ダブルクォート終わりまで飛ばす
				if (line[0] == '"')
				{
					// 後ろのダブルクォートのIndexを取得
					int index = line.IndexOf('"', 1) + 1;
					if (index >= 0)
					{
						if (index < minIndex)
						{
							minIndex = index;
						}
					}
				}
				else
				{
					// 分割指定されたCharが最初に出てくるIndexを取得
					foreach (char c in splitChars)
					{
						int index = line.IndexOf(c);
						if (index >= 0)
						{
							if (index < minIndex)
							{
								minIndex = index;
							}
						}
					}
				}
				if (minIndex == int.MaxValue)
				{
					// 見つからなかった場合トークンを抜き出し、テキストを抽出して終わり
					token.Add(line);
					break;
				}
				else
				{
					// トークンを抽出し
					string tok = line.Substring(0, minIndex);
					token.Add(tok);
					// テキストを短縮
					int tokLength = tok.Length + 1;
					if (tokLength < line.Length)
					{
						string newLine = line.Substring(tokLength);
						line = newLine;
					}
					else
					{
						break;
					}
				}
			}
			return token.ToArray();
		}

		/// <summary>
		/// ダブルクォートで括ったファイル名文字列を、トリムして取得。
		/// </summary>
		/// <param name="text">テキスト</param>
		/// <returns>テキスト</returns>
		private string TrimQuotedString(string text)
		{
			string result = string.Empty;
			if(!string.IsNullOrEmpty(text))
			{
				result = text.Trim(TRIM_CHAR);
			}
			return result;
		}

		#endregion private

	}

	#endregion NESP ProcessData

	#region NESP Exception

	/// <summary>
	/// 例外。スクリプトの行を参照可能。
	/// </summary>
	public class NespException : Exception
	{
		public string ScriptLine { get; set; }
		public bool BadCommand { get; set; }
		public bool BadInitialForCommand { get; set; }

		public NespException(string scriptLine)
		{
			ScriptLine = scriptLine;
		}

		public NespException(Exception e, string scriptLine)
			: base(e.Message, e)
		{
			ScriptLine = scriptLine;
		}

	}

	#endregion NESP Exception

	#region NESP

	/// <summary>
	/// NESPスクリプト実行クラス。
	/// </summary>
	public class Nesp
	{

		public event EventHandler FileRequired = null;

		private string mScriptPath = string.Empty;
		private NespProcess[] mScriptProcess = null;

		private Dictionary<string, byte[]> mSrcFiles = new Dictionary<string, byte[]>();

		#region Constructor

		/// <summary>
		/// コンストラクタ。
		/// </summary>
		/// <param name="scriptPath">スクリプトファイルの場所</param>
		/// <param name="script">スクリプト</param>
		public Nesp(string scriptPath, string script)
		{
			mScriptPath = scriptPath;
			mScriptProcess = ConvertScript(script);
		}

		/// <summary>
		/// コンストラクタ。
		/// </summary>
		/// <param name="scriptPath">スクリプトファイルの場所</param>
		/// <param name="script">スクリプト</param>
		/// <param name="targetFiles">対象ファイル配列</param>
		public Nesp(string scriptPath, string script, string[] targetFiles)
		{
			string targetScript = "";
			for (int i = 0; i < targetFiles.Length; i++)
			{
				string targetFile = targetFiles[i];
				if (i == 0)
				{
					targetScript += ("SET_FILE \"" + targetFile + "\"\n");
				}
				char initial = (char)('A' + i);
				targetScript += (initial + ".SET_FILE \"" + targetFile + "\"\n");
			}
			script = targetScript + script;

			mScriptPath = scriptPath;
			mScriptProcess = ConvertScript(script);
		}

		/// <summary>
		/// コンストラクタ。
		/// </summary>
		/// <param name="scriptPath">スクリプトファイルの場所</param>
		/// <param name="script">スクリプト</param>
		/// <param name="targetPath">対象ファイル</param>
		public Nesp(string scriptPath, string script, string targetPath)
			: this(scriptPath, script, new string[] { targetPath })
		{
		}

		#endregion Constructor

		#region Convert Script

		/// <summary>
		/// スクリプトから処理リストに変換
		/// </summary>
		/// <param name="script">スクリプト</param>
		/// <returns>処理リスト</returns>
		private NespProcess[] ConvertScript(string script)
		{
			List<NespProcess> scriptProcessList = new List<NespProcess>();

			// 全ての行を解析
			StringReader sr = new StringReader(script);
			string line = sr.ReadLine();
			while (line != null)
			{
				// 空行を無視
				if (string.IsNullOrEmpty(line))
				{
					// 次の行を読み込む
					line = sr.ReadLine();
					continue;
				}
				// コメント行を無視
				if (line.StartsWith(";"))
				{
					// 次の行を読み込む
					line = sr.ReadLine();
					continue;
				}
				
				// 行の解析
				NespProcess process = new NespProcess(line);
				scriptProcessList.Add(process);

				// 次の行を読み込む
				line = sr.ReadLine();
			}

			return scriptProcessList.ToArray();
		}

		#endregion Convert Script

		#region Execute Script

		#region Execute

		/// <summary>
		/// スクリプト実行。
		/// </summary>
		/// <returns>出力データ</returns>
		public byte[] Execute()
		{
			List<byte> data = new List<byte>();

			// 全ての行を実行
			foreach (NespProcess process in mScriptProcess)
			{
				try
				{
					StepProcess(process, data);
				}
				catch(IpsException)
				{
					throw;
				}
				catch
				{
					throw new NespException(process.Line);
				}
			}

			return data.ToArray();
		}

		#endregion Execute

		#region Step Process

		/// <summary>
		/// 処理の実行。
		/// </summary>
		/// <param name="process">処理</param>
		/// <param name="data">出力データ</param>
		private void StepProcess(NespProcess process, List<byte> data)
		{
			switch (process.ProcessType)
			{
				// IPSパッチ
				case ProcessType.Ips:
					{
						string filename;
						if (string.IsNullOrEmpty(process.Filename))
						{
							filename = Path.ChangeExtension(mScriptPath, "ips");
						}
						else
						{
							filename = GetFullPathFromScriptPath(process.Filename);
						}
						Patch(filename, data);
					}
					break;

				// 外部ファイル
				case ProcessType.AddFile:
					{
						string filename = GetFullPathFromScriptPath(process.Filename);
						byte[] fileData = File.ReadAllBytes(filename);
						data.AddRange(fileData);
					}
					break;

				// ダミーデータ
				case ProcessType.AddDummy:
					{
						byte[] dummyData = new byte[process.Size];
						for (int i = 0; i < dummyData.Length; i++)
						{
							dummyData[i] = process.ByteValue;
						}
						data.AddRange(dummyData);
					}
					break;

				// ファイル書き出し
				case ProcessType.WriteFile:
					{
						WriteFile(process.Filename, data);
					}
					break;

				// ファイル指定
				case ProcessType.SetFile:
					{
						string initial = process.Initial;
						if (mSrcFiles.ContainsKey(initial))
						{
							mSrcFiles.Remove(initial);
						}
						string filename = GetFullPathFromScriptPath(process.Filename);
						byte[] filedata = File.ReadAllBytes(filename);
						mSrcFiles.Add(initial, filedata);
					}
					break;

				// サイズチェック
				case ProcessType.CheckSize:
					{
						if (mSrcFiles.ContainsKey(process.Initial))
						{
							byte[] srcData = mSrcFiles[process.Initial];
							if (srcData.Length != process.Size)
							{
								throw new Exception("CHECK_SIZE error");
							}
						}
						else
						{
							throw new Exception("Source File not selected error");
						}
					}
					break;

				// 結合処理
				case ProcessType.Add:
					{
						if (mSrcFiles.ContainsKey(process.Initial))
						{
							byte[] srcData = mSrcFiles[process.Initial];
							NesHeader header = new NesHeader(srcData);

							int addr = process.Addr;
							// PRG-ROMなら、NESヘッダ分のアドレスを足す
							if (process.Command.StartsWith("PRG"))
							{
								addr += 0x000010;
							}
							// CHR-ROMなら、NESヘッダ+PRG-ROM分のアドレスを足す
							if (process.Command.StartsWith("CHR"))
							{
								int prgSize = header.PrgRom * 0x004000;
								addr += (prgSize + 0x000010);
							}
							//AddData(data, mTarget, addr, process.Size);
							AddData(data, srcData, addr, process.Size);
						}
						else
						{
							throw new Exception("Source File not selected error");
						}
					}
					break;

				case ProcessType.None:
				default:
					break;
			}
		}

		#endregion Step Process

		#endregion Execute Script

		#region Process

		#region ADD

		/// <summary>
		/// 元データのアドレスからサイズ分、追加先データに追加する。
		/// </summary>
		/// <param name="dst">追加先データ</param>
		/// <param name="src">元データ</param>
		/// <param name="srcAddr">元データアドレス</param>
		/// <param name="srcSize">元データサイズ</param>
		private void AddData(List<byte> dst, byte[] src, int srcAddr, int srcSize)
		{
			for (int i = 0; i < srcSize; i++)
			{
				int addr = srcAddr + i;
				if (addr < src.Length)
				{
					dst.Add(src[addr]);
				}
				else
				{
					break;
				}
			}
		}

		#endregion ADD

		#region IPS

		/// <summary>
		/// パッチ処理。
		/// </summary>
		/// <param name="filename">パッチファイル名</param>
		/// <param name="data">データ</param>
		private void Patch(string filename, List<byte> data)
		{
			byte[] patchData = File.ReadAllBytes(filename);
			Ips.Patch(data, patchData);
		}

		#endregion IPS

		#region WriteFile

		/// <summary>
		/// ファイル書き出し。
		/// </summary>
		/// <param name="filename">出力ファイル名</param>
		/// <param name="data">データ</param>
		private void WriteFile(string filename, List<byte> data)
		{
			if (data != null)
			{
				byte[] byteData = data.ToArray();
				File.WriteAllBytes(filename, byteData);
			}
		}

		#endregion WriteFile

		#endregion Process

		#region Common

		/// <summary>
		/// スクリプトファイルのパスの、相対ファイル名を指定して、絶対パスを取得する
		/// </summary>
		/// <param name="filename">相対ファイル名</param>
		/// <returns>絶対パス</returns>
		private string GetFullPathFromScriptPath(string filename)
		{
			string newFilename = filename;
			if (!File.Exists(filename))
			{
				newFilename = Path.GetDirectoryName(mScriptPath) + "\\" + filename;
			}
			return newFilename;
		}

		#endregion Common

	}

	#endregion NESP

}
