﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace NesExtendPatchTool
{
	public partial class MainForm : Form
	{

		const string EXT_PATCH = "nesp";
		const string IPS_PATCH = "ips";

		private bool mIsNesp = false;
		private bool mIsIps = false;

		public MainForm()
		{
			InitializeComponent();
		}

		#region ドラッグアンドドロップ

		/// <summary>
		/// ファイルをフォーム内にドラッグしたときの処理。
		/// ・ドラッグアイテム判定とカーソルの変更。
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void hDragEnter(object sender, DragEventArgs e)
		{
			if (e.Data.GetDataPresent(DataFormats.FileDrop))
			{
				//ドラッグされたデータ形式を調べ、ファイルのときはコピーカーソルとする
				e.Effect = DragDropEffects.All;
			}
			else
			{
				//ファイル以外は受け付けない
				e.Effect = DragDropEffects.None;
			}
		}

		/// <summary>
		/// ファイルをドロップした時の処理。
		/// ・ファイル数が1個なら自動判定でファイル名を展開する。
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void hDragDrop(object sender, DragEventArgs e)
		{
			string[] fileName = (string[])e.Data.GetData(DataFormats.FileDrop, false);
			if (fileName.Length == 1)
			{
				string filename = fileName[0];
				string fileExt = Path.GetExtension(filename);
				if (!string.IsNullOrEmpty(fileExt))
				{
					if (fileExt.ToLower().Contains(EXT_PATCH))
					{
						// パッチのファイル名を設定
						SetScriptFileName(filename);
					}
					else if (fileExt.ToLower().Contains(IPS_PATCH))
					{
						// パッチのファイル名を設定
						SetScriptFileName(filename);
					}
					else
					{
						// 対象ファイルのファイル名を設定
						SetTargetFileName(filename);
					}
				}
			}
		}

		#endregion ドラッグアンドドロップ

		#region ボタン

		private void buttonOpenScript_Click(object sender, EventArgs ev)
		{
			if(openFileDialogPatch.ShowDialog()==DialogResult.OK)
			{
				string filename = openFileDialogPatch.FileName;
				// パッチのファイル名を設定
				SetScriptFileName(filename);
			}
		}

		private void buttonOpenTarget_Click(object sender, EventArgs ev)
		{
			if (openFileDialogTarget.ShowDialog() == DialogResult.OK)
			{
				string filename = openFileDialogTarget.FileName;
				// 対象ファイルのファイル名を設定
				SetTargetFileName(filename);
			}
		}

		private void buttonExecute_Click(object sender, EventArgs ev)
		{
			byte[] saveData = null;
			string saveFilename = textBoxSaveFile.Text;
			if (string.IsNullOrEmpty(saveFilename))
			{
				MessageBox.Show("出力ファイル名を入力してください。", "Error");
				return;
			}

			// 上書きチェック
			if (File.Exists(saveFilename))
			{
				if (MessageBox.Show("出力先に既にファイルが存在します。\n上書きしますか？", "上書き確認", MessageBoxButtons.YesNo, MessageBoxIcon.Question) != DialogResult.Yes)
				{
					return;
				}
			}

			// ファイルを開く
			string targetFilename = textBoxTargetFilename.Text;
			byte[] targetData = null;
			try
			{
				targetData = File.ReadAllBytes(targetFilename);
			}
			catch (Exception e)
			{
				MessageBox.Show("対象ファイルを開けませんでした。\n" + e.Message, "Error");
				return;
			}

			// パッチ実行
			string patchFilename = textBoxScriptFilename.Text;
			if (mIsNesp)
			{
				// スクリプト取得
				string script = textBoxScript.Text;
				/*
				if (targetFilename != string.Empty)
				{
					script += ("SET_FILE \"" + targetFilename + "\"\n");
					script += ("A.SET_FILE \"" + targetFilename + "\"\n");
				}
				 */
				// スクリプトを処理
				try
				{
					// 分割/結合
					Nesp nesp = new Nesp(patchFilename, script, targetFilename);
					saveData = nesp.Execute();
				}
				catch (NespException e)
				{
					if(e.BadCommand)
					{
						MessageBox.Show("NESPスクリプトの処理に失敗しました。\nサポートされていないコマンドです。\n" + e.ScriptLine, "Error");
					}
					else if(e.BadInitialForCommand)
					{
						MessageBox.Show("NESPスクリプトの処理に失敗しました。\nイニシャルを指定できないコマンドで、イニシャルが指定されています。\n" + e.ScriptLine, "Error");
					}
					else
					{
						MessageBox.Show("NESPスクリプトの処理に失敗しました。\n" + e.ScriptLine, "Error");
					}
					return;
				}
				catch (IpsException e)
				{
					MessageBox.Show("IPSパッチに失敗しました。[0]\n" + e.Message, "Error");
					return;
				}
				catch (Exception e)
				{
					MessageBox.Show("分割・結合に失敗しました。\n" + e.Message, "Error");
					return;
				}
			}
			else if (mIsIps)
			{
				try
				{
					byte[] patchData = File.ReadAllBytes(patchFilename);
					List<byte> data = targetData.ToList<byte>();
					Ips.Patch(data, patchData);
					saveData = data.ToArray();
				}
				catch (IpsException e)
				{
					MessageBox.Show("IPSパッチに失敗しました。[1]\n" + e.Message, "Error");
					return;
				}
				catch (Exception e)
				{
					MessageBox.Show("IPSパッチに失敗しました。[2]\n" + e.Message, "Error");
					return;
				}
			}
			else
			{
				MessageBox.Show("パッチファイルが正しくありません。", "Error");
				return;
			}

			try
			{
				File.WriteAllBytes(saveFilename, saveData);
			}
			catch (Exception e)
			{
				MessageBox.Show("ファイルに保存できませんでした。\n" + e.Message, "Error");
				return;
			}

			MessageBox.Show("処理完了。", "Information");
		}

		private void buttonSaveScript_Click(object sender, EventArgs e)
		{
			// overwrite script
		}

		#endregion ボタン

		#region 処理

		/// <summary>
		/// 対象ファイルのファイル名を設定
		/// </summary>
		/// <param name="filename"></param>
		private void SetTargetFileName(string filename)
		{
			// 対象ファイルのファイル名を設定
			textBoxTargetFilename.Text = filename;

			// 出力ファイル名を更新
			string patchFilename = textBoxScriptFilename.Text;
			if (!string.IsNullOrEmpty(patchFilename))
			{
				UpdateSaveFileName(patchFilename, filename);
			}
		}

		/// <summary>
		/// パッチのファイル名を設定
		/// </summary>
		/// <param name="filename"></param>
		private void SetScriptFileName(string filename)
		{
			// パッチのファイル名を設定
			textBoxScriptFilename.Text = filename;

			mIsNesp = false;
			mIsIps = false;

			string fileExt = Path.GetExtension(filename);
			if (!string.IsNullOrEmpty(fileExt))
			{
				if (fileExt.ToLower().Contains(EXT_PATCH))
				{
					mIsNesp = true;

					// スクリプトの読み込み
					ReadScript();
				}
				else if (fileExt.ToLower().Contains(IPS_PATCH))
				{
					mIsIps = true;

					// スクリプトを消去
					textBoxScript.Text = string.Empty;
				}
			}

			// 出力ファイル名を更新
			string targetFilename = textBoxTargetFilename.Text;
			if (!string.IsNullOrEmpty(targetFilename))
			{
				UpdateSaveFileName(filename, targetFilename);
			}
		}

		/// <summary>
		/// スクリプトの読み込み
		/// </summary>
		private void ReadScript()
		{
			string script = string.Empty;
			try
			{
				string filename = textBoxScriptFilename.Text;
				//script = File.ReadAllText(filename);
				byte[] scriptByteFile = File.ReadAllBytes(filename);
				Encoding enc = GetEncoding.GetCode(scriptByteFile);
				if (enc != null)
				{
					using (StreamReader sr = new StreamReader(filename, enc))
					{
						//内容をすべて読み込む
						script = sr.ReadToEnd();
						//閉じる
						sr.Close();
					}
				}
				else
				{
					throw new Exception("Bad Script Encoding.");
				}
			}
			catch (Exception e)
			{
				MessageBox.Show("スクリプトファイルが正しくありません。\n" + e.Message, "Error");
			}
			textBoxScript.Text = script;
		}

		/// <summary>
		/// 出力ファイル名を更新
		/// </summary>
		/// <param name="patchFilename"></param>
		/// <param name="filename"></param>
		private void UpdateSaveFileName(string patchFilename, string targetFilename)
		{
			if ((string.IsNullOrEmpty(patchFilename)) || (string.IsNullOrEmpty(targetFilename))) return;

			try
			{
				string fileExt = Path.GetExtension(targetFilename);
				string filename = Path.ChangeExtension(patchFilename, fileExt);

				textBoxSaveFile.Text = filename;
			}
			catch
			{
			}
		}

		#endregion 処理


	}
}
